@@ -4,11 +4,11 @@
  * SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
  */
 #include "decoration.h"
-#include "decoratedclient.h"
+#include "decoratedwindow.h"
 #include "decoration_p.h"
 #include "decorationbutton.h"
 #include "decorationsettings.h"
-#include "private/decoratedclientprivate.h"
+#include "private/decoratedwindowprivate.h"
 #include "private/decorationbridge.h"
 
 #include <QCoreApplication>
@@ -16,7 +16,7 @@
 
 #include <cmath>
 
-namespace KDecoration2
+namespace KDecoration3
 {
 namespace
 {
@@ -31,10 +31,45 @@ DecorationBridge *findBridge(const QVariantList &args)
 }
 }
 
+class DecorationStateData : public QSharedData
+{
+public:
+    QMarginsF borders;
+};
+
+DecorationState::DecorationState()
+    : d(new DecorationStateData)
+{
+}
+
+DecorationState::DecorationState(const DecorationState &other)
+    : d(other.d)
+{
+}
+
+DecorationState::~DecorationState()
+{
+}
+
+std::shared_ptr<DecorationState> DecorationState::clone() const
+{
+    return std::make_shared<DecorationState>(*this);
+}
+
+QMarginsF DecorationState::borders() const
+{
+    return d->borders;
+}
+
+void DecorationState::setBorders(const QMarginsF &borders)
+{
+    d->borders = borders;
+}
+
 Decoration::Private::Private(Decoration *deco, const QVariantList &args)
     : sectionUnderMouse(Qt::NoSection)
     , bridge(findBridge(args))
-    , client(std::shared_ptr<DecoratedClient>(new DecoratedClient(deco, bridge)))
+    , client(std::shared_ptr<DecoratedWindow>(new DecoratedWindow(deco, bridge)))
     , opaque(false)
     , q(deco)
 {
@@ -51,20 +86,21 @@ void Decoration::Private::setSectionUnderMouse(Qt::WindowFrameSection section)
 
 void Decoration::Private::updateSectionUnderMouse(const QPoint &mousePosition)
 {
-    if (titleBar.contains(mousePosition)) {
+    if (titleBar.toRect().contains(mousePosition)) {
         setSectionUnderMouse(Qt::TitleBarArea);
         return;
     }
-    const QSize size = q->size();
+    const QSizeF size = q->size();
+    const QMarginsF borders = current->borders();
     const int corner = 2 * settings->largeSpacing();
     const bool left = mousePosition.x() < borders.left();
     const bool top = mousePosition.y() < borders.top();
-    const bool bottom = size.height() - mousePosition.y() <= borders.bottom();
-    const bool right = size.width() - mousePosition.x() <= borders.right();
+    const bool bottom = mousePosition.y() >= size.height() - borders.bottom();
+    const bool right = mousePosition.x() >= size.width() - borders.right();
     if (left) {
         if (top && mousePosition.y() < titleBar.top() + corner) {
             setSectionUnderMouse(Qt::TopLeftSection);
-        } else if (size.height() - mousePosition.y() <= borders.bottom() + corner && mousePosition.y() > titleBar.bottom()) {
+        } else if (mousePosition.y() >= size.height() - borders.bottom() - corner && mousePosition.y() >= titleBar.bottom()) {
             setSectionUnderMouse(Qt::BottomLeftSection);
         } else {
             setSectionUnderMouse(Qt::LeftSection);
@@ -74,7 +110,7 @@ void Decoration::Private::updateSectionUnderMouse(const QPoint &mousePosition)
     if (right) {
         if (top && mousePosition.y() < titleBar.top() + corner) {
             setSectionUnderMouse(Qt::TopRightSection);
-        } else if (size.height() - mousePosition.y() <= borders.bottom() + corner && mousePosition.y() > titleBar.bottom()) {
+        } else if (mousePosition.y() >= size.height() - borders.bottom() - corner && mousePosition.y() >= titleBar.bottom()) {
             setSectionUnderMouse(Qt::BottomRightSection);
         } else {
             setSectionUnderMouse(Qt::RightSection);
@@ -82,10 +118,10 @@ void Decoration::Private::updateSectionUnderMouse(const QPoint &mousePosition)
         return;
     }
     if (bottom) {
-        if (mousePosition.y() > titleBar.bottom()) {
+        if (mousePosition.y() >= titleBar.bottom()) {
             if (mousePosition.x() < borders.left() + corner) {
                 setSectionUnderMouse(Qt::BottomLeftSection);
-            } else if (size.width() - mousePosition.x() <= borders.right() + corner) {
+            } else if (mousePosition.x() >= size.width() - borders.right() - corner) {
                 setSectionUnderMouse(Qt::BottomRightSection);
             } else {
                 setSectionUnderMouse(Qt::BottomSection);
@@ -99,7 +135,7 @@ void Decoration::Private::updateSectionUnderMouse(const QPoint &mousePosition)
         if (mousePosition.y() < titleBar.top()) {
             if (mousePosition.x() < borders.left() + corner) {
                 setSectionUnderMouse(Qt::TopLeftSection);
-            } else if (size.width() - mousePosition.x() <= borders.right() + corner) {
+            } else if (mousePosition.x() >= size.width() - borders.right() - corner) {
                 setSectionUnderMouse(Qt::TopRightSection);
             } else {
                 setSectionUnderMouse(Qt::TopSection);
@@ -132,14 +168,11 @@ Decoration::Decoration(QObject *parent, const QVariantList &args)
     : QObject(parent)
     , d(new Private(this, args))
 {
-    connect(this, &Decoration::bordersChanged, this, [this] {
-        update();
-    });
 }
 
 Decoration::~Decoration() = default;
 
-DecoratedClient *Decoration::client() const
+DecoratedWindow *Decoration::window() const
 {
     return d->client.get();
 }
@@ -179,7 +212,7 @@ void Decoration::requestToggleKeepBelow()
     d->client->d->requestToggleKeepBelow();
 }
 
-#if KDECORATIONS2_ENABLE_DEPRECATED_SINCE(5, 21)
+#if KDECORATIONS3_ENABLE_DEPRECATED_SINCE(5, 21)
 void Decoration::requestShowWindowMenu()
 {
     requestShowWindowMenu(QRect());
@@ -218,9 +251,7 @@ void Decoration::showApplicationMenu(int actionId)
 
 void Decoration::requestShowApplicationMenu(const QRect &rect, int actionId)
 {
-    if (auto *appMenuEnabledPrivate = dynamic_cast<ApplicationMenuEnabledDecoratedClientPrivate *>(d->client->d.get())) {
-        appMenuEnabledPrivate->requestShowApplicationMenu(rect, actionId);
-    }
+    d->client->d->requestShowApplicationMenu(rect, actionId);
 }
 
 void Decoration::setBlurRegion(const QRegion &region)
@@ -231,15 +262,16 @@ void Decoration::setBlurRegion(const QRegion &region)
     }
 }
 
-void Decoration::setBorders(const QMargins &borders)
+void Decoration::setBorders(const QMarginsF &borders)
 {
-    if (d->borders != borders) {
-        d->borders = borders;
-        Q_EMIT bordersChanged();
+    if (d->next->borders() != borders) {
+        setState([borders](DecorationState *state) {
+            state->setBorders(borders);
+        });
     }
 }
 
-void Decoration::setResizeOnlyBorders(const QMargins &borders)
+void Decoration::setResizeOnlyBorders(const QMarginsF &borders)
 {
     if (d->resizeOnlyBorders != borders) {
         d->resizeOnlyBorders = borders;
@@ -247,7 +279,7 @@ void Decoration::setResizeOnlyBorders(const QMargins &borders)
     }
 }
 
-void Decoration::setTitleBar(const QRect &rect)
+void Decoration::setTitleBar(const QRectF &rect)
 {
     if (d->titleBar != rect) {
         d->titleBar = rect;
@@ -276,17 +308,17 @@ QRegion Decoration::blurRegion() const
     return d->blurRegion;
 }
 
-QMargins Decoration::borders() const
+QMarginsF Decoration::borders() const
 {
-    return d->borders;
+    return d->current->borders();
 }
 
-QMargins Decoration::resizeOnlyBorders() const
+QMarginsF Decoration::resizeOnlyBorders() const
 {
     return d->resizeOnlyBorders;
 }
 
-QRect Decoration::titleBar() const
+QRectF Decoration::titleBar() const
 {
     return d->titleBar;
 }
@@ -306,56 +338,55 @@ bool Decoration::isOpaque() const
     return d->opaque;
 }
 
-int Decoration::borderLeft() const
+qreal Decoration::borderLeft() const
 {
-    return d->borders.left();
+    return d->current->borders().left();
 }
 
-int Decoration::resizeOnlyBorderLeft() const
+qreal Decoration::resizeOnlyBorderLeft() const
 {
     return d->resizeOnlyBorders.left();
 }
 
-int Decoration::borderRight() const
+qreal Decoration::borderRight() const
 {
-    return d->borders.right();
+    return d->current->borders().right();
 }
 
-int Decoration::resizeOnlyBorderRight() const
+qreal Decoration::resizeOnlyBorderRight() const
 {
     return d->resizeOnlyBorders.right();
 }
 
-int Decoration::borderTop() const
+qreal Decoration::borderTop() const
 {
-    return d->borders.top();
+    return d->current->borders().top();
 }
 
-int Decoration::resizeOnlyBorderTop() const
+qreal Decoration::resizeOnlyBorderTop() const
 {
     return d->resizeOnlyBorders.top();
 }
 
-int Decoration::borderBottom() const
+qreal Decoration::borderBottom() const
 {
-    return d->borders.bottom();
+    return d->current->borders().bottom();
 }
 
-int Decoration::resizeOnlyBorderBottom() const
+qreal Decoration::resizeOnlyBorderBottom() const
 {
     return d->resizeOnlyBorders.bottom();
 }
 
-QSize Decoration::size() const
+QSizeF Decoration::size() const
 {
-    const QMargins &b = d->borders;
-    return QSize(d->client->width() + b.left() + b.right(), //
-                 (d->client->isShaded() ? 0 : d->client->height()) + b.top() + b.bottom());
+    const QMarginsF b = d->current->borders();
+    return QSizeF(d->client->width() + b.left() + b.right(), (d->client->isShaded() ? 0 : d->client->height()) + b.top() + b.bottom());
 }
 
-QRect Decoration::rect() const
+QRectF Decoration::rect() const
 {
-    return QRect(QPoint(0, 0), size());
+    return QRectF(QPointF(0, 0), size());
 }
 
 bool Decoration::event(QEvent *event)
@@ -472,9 +503,9 @@ void Decoration::wheelEvent(QWheelEvent *event)
     }
 }
 
-void Decoration::update(const QRect &r)
+void Decoration::update(const QRectF &r)
 {
-    Q_EMIT damaged(r.isNull() ? rect() : r);
+    Q_EMIT damaged(r.isNull() ? rect().toAlignedRect() : r.toAlignedRect());
 }
 
 void Decoration::update()
@@ -492,6 +523,50 @@ std::shared_ptr<DecorationSettings> Decoration::settings() const
     return d->settings;
 }
 
+std::shared_ptr<DecorationState> Decoration::createState()
+{
+    return std::make_shared<DecorationState>();
+}
+
+std::shared_ptr<DecorationState> Decoration::currentState() const
+{
+    return d->current;
+}
+
+std::shared_ptr<DecorationState> Decoration::nextState() const
+{
+    return d->next;
+}
+
+void Decoration::create()
+{
+    d->next = createState();
+    d->current = createState();
+}
+
+void Decoration::setState(std::function<void(DecorationState *state)> callback)
+{
+    callback(d->next.get());
+    Q_EMIT nextStateChanged(d->next);
+}
+
+void Decoration::apply(std::shared_ptr<DecorationState> state)
+{
+    if (d->current == state) {
+        return;
+    }
+
+    const auto previous = d->current;
+    d->current = state;
+    update();
+
+    if (previous->borders() != state->borders()) {
+        Q_EMIT bordersChanged();
+    }
+
+    Q_EMIT currentStateChanged(state);
+}
+
 } // namespace
 
 #include "moc_decoration.cpp"
